r"""GNN profiling package."""

from .profile import profileit, timeit, get_stats_summary
from .profile import (
    trace_handler,
    print_time_total,
    rename_profile_file,
    torch_profile,
    xpu_profile,
)
from .utils import count_parameters
from .utils import get_model_size
from .utils import get_data_size
from .utils import get_cpu_memory_from_gc
from .utils import get_gpu_memory_from_gc
from .utils import get_gpu_memory_from_nvidia_smi
from .benchmark import benchmark

__all__ = [
    'profileit',
    'timeit',
    'get_stats_summary',
    'trace_handler',
    'print_time_total',
    'rename_profile_file',
    'torch_profile',
    'xpu_profile',
    'count_parameters',
    'get_model_size',
    'get_data_size',
    'get_cpu_memory_from_gc',
    'get_gpu_memory_from_gc',
    'get_gpu_memory_from_nvidia_smi',
    'benchmark',
]

classes = __all__
